import nbformat as nbf
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import CellExecutionError
import pandas as pd
import sys
import os
from os import path
import shutil
from nbta.utils import footer_cell, header_cell, header_code_cell
from tqdm import tqdm


class ParsedNotebook():
    '''
    Usage: ParsedNotebook(path, questions)

    A parsed Jupyter notebook and associated questions
    '''

    def __init__(self, path, author, file_name):
        with open(path, 'r') as f:
            self.content = nbf.read(f, as_version=nbf.NO_CONVERT)
        base_dir = '/'.join(path.split('/')[:-1])
        self.file_name = f'{base_dir}/{file_name}'
        self.author = author
        self.modified_content = None

    def insert_cells(self, new_cells):
        cells = self.content['cells']
        self.modified_content = self.content.copy()
        modified_cells = header_cell(self.author) + [header_code_cell(self.author)]

        for this_cell in cells:
            inserted = False
            cells_before = []
            cells_after = []
            begin_marking = '<span style="color:darkred">==============================================================MARK THIS QUESTION BELOW=================================================================</span>\n<h1 style="color:darkred">YOUR MARKS</h1>'
            for new_cell in new_cells:
                if  new_cell.tag in this_cell['source']:
                    if new_cell.position == 'before':
                        if len(cells_before)==0:
                            cells_before.append(nbf.v4.new_markdown_cell(source=begin_marking))
                        cells_before.append(new_cell.cell)
                    else:
                        if len(cells_after)==0:
                            cells_after.append(nbf.v4.new_markdown_cell(source=begin_marking)) 
                        cells_after.append(new_cell.cell)
                    inserted = True
            if inserted:
                cells_before.append(this_cell)
                modified_cells = modified_cells + cells_before + cells_after
            else:
                modified_cells.append(this_cell)

        modified_cells.append(nbf.v4.new_markdown_cell(source='<h1 style="color:blue">Coding style and private feedback to lecturer</h1>"'))
        modified_cells.append(nbf.v4.new_code_cell(source=footer_cell()))
        tests_list = [f'nbta_test_{c.name}' for c in new_cells]
        tests_list=tests_list+['nbta_test_style','nbta_estimated_mark','nbta_private_feedback']
        title_cell_text = '<h1 style="color:red">RUN the cell below to save your markings</h1>"'
        modified_cells.append(nbf.v4.new_markdown_cell(source=title_cell_text))
        final_cell_text = ','.join(tests_list)
        final_cell_code = ['# DONT FORGET TO SAVE:','',f'import os','',
        f'all_tests = [{final_cell_text}]','',
        f'for t in all_tests:',
        f'    t.save_values()','',
        f'print("Saved First Marker as:", os.getlogin())'
        f'pd.Series(data=[os.getlogin()], name="marker_name").to_csv("grades/first_marker.csv", index=False)']
        final_cell_code = '\n'.join(final_cell_code)
        modified_cells.append(nbf.v4.new_code_cell(source=final_cell_code))

        self.modified_content['cells'] = modified_cells

        return self

    def yield_question_cells(self,start_question,end_question):
        return_cells = [nbf.v4.new_markdown_cell(source=f'ANSWER FROM: {self.author}')]
        in_scope = False
        for this_cell in self.content['cells']:
            if start_question.tags in this_cell['source']:
                in_scope = True
                return_cells.append(this_cell)
            if end_question is not None:
                if end_question.tags in this_cell['source']:
                    in_scope = False
                    return return_cells
            if in_scope:
                return_cells.append(this_cell)
        return return_cells

    def execute_notebook(self, kernel):
        ep = ExecutePreprocessor(timeout=99999999,kernel_name=kernel)
        ep.preprocess(self.modified_content)

        return None

    def write(self):
        if self.modified_content is None:
            content = self.content
        else:
            content = self.modified_content

        with open(f'{self.file_name}.ipynb', 'w') as f:
            nbf.write(content,f)

class MarkingCell():
    def __init__(self, name, tag, cell_type='code', position='before', from_file=True, source_data=None):
        if from_file is False and source_data is None:
            raise Exception("Either from_file must be True or source_data must not be a NoneType")

        if from_file is True:
            self.check_source_dir()
            source_path = f"grading/testing/notebook_tests/{name}.py"
            if not os.path.exists(source_path):
                self.create_test_file(source_path)
            
            with open(source_path, 'r') as f:
                source = ''.join(f.readlines())
            
            options_path = f"grading/testing/notebook_tests/{name}.csv"
            if not os.path.exists(options_path):
                self.initialise_options(options_path)
            
            style_path = f"grading/testing/notebook_tests/style.csv"
            if not os.path.exists(style_path):
                self.initialise_styles(style_path)

        if cell_type == 'code':
            self.cell = nbf.v4.new_code_cell(source=source)
        elif cell_type == 'markdown':
            self.cell = nbf.v4.new_markdown_cell(source=source)
        else:
            raise Exception(f"Unknown cell type: {cell_type}")
        self.tag = tag
        self.position = position
        self.name = name

    def check_source_dir(self):
        grading_dirs = ["grading", "grading/testing","grading/testing/notebook_tests",
        "grading/scores","grading/scores/individual", "grading/testing/external_tests"]

        for path in grading_dirs:
            if not os.path.isdir(path):
                os.mkdir(path)

    def create_test_file(self, path):
        test_name = path.split("/")[-1].strip(".py")
        code = [f"# TEST FOR QUESTION {test_name}",
        f"nbta_test_{test_name} = QuestionGrader('{test_name}')"]
        code = "\n".join(code)

        with open(path, 'w') as f:
                f.write(code)

    def initialise_options(self, path):
        options = "\n".join(["options,feedback","not_answered,you have not answered the question", 
        "failed, you have not understood this question", "pass, you understood the basics of this question; However more could have been done.", 
        "high_pass, you understood the basics of this question", "merit, you understood the question well","high_merit, you answered the question very well", "distinction, your answer is in the top 10% of the class", "high_distinction, your answer is in the top 5% of the class"])
        with open(path, 'w') as f:
                f.write(options)

    def initialise_styles(self, path):
        options = "\n".join(["options,feedback",
        "good use of functions","logical variable names","pythonic style","good use of markdown",
        "appropriate level of comments", "good use of graphs", "some graphs", "no graphs",
        "best solution", "simple solution", "complicated solution", "overly complicated solution"])
        with open(path, 'w') as f:
                f.write(options)
 
class FeedbackCell:
    def __init__(self, tag, name="STUDENT_FEEDBACK",position='before'):
        self.name = name
        self.tag = tag
        self.position = position

        source = f"nbta_test_{self.name} = QuestionGrader('{self.name}', feedback=True)"
        self.cell = nbf.v4.new_code_cell(source=source)
        return None

class NotebookMarker():
    def __init__(self, folder, notebook_name, questions=None,name_func=None):
        self.notebook_name = notebook_name
        self.name_func = name_func
        self.base_dir = folder
        self.marking_name = f'{self.notebook_name}_marking'
        self.questions = questions
        self.candidates = self.filter_dirs(os.listdir(self.base_dir))
        self.notebooks = self.get_notebooks()
        self.test_list = None

    def filter_dirs(self, candidates):
        if '.DS_Store' in candidates:
            candidates.remove('.DS_Store')
        if '.ipynb_checkpoints' in candidates:
            candidates.remove('.ipynb_checkpoints')
        return candidates

    def insert_cells(self, cells_data=None):
        if cells_data is None:
            if self.questions is None:
                return self
            cells_data = self.questions

        for auth, notebook in self.notebooks.items():
            notebook.insert_cells(cells_data).write()
        return self

    def get_notebooks(self):
        notebooks = {}
        for candidate in self.candidates:
            try:
                path = f'{self.base_dir}/{candidate}/{self.notebook_name}.ipynb'
                notebook = ParsedNotebook(path, candidate, self.marking_name)
                notebooks[candidate] = notebook
            except Exception as e:
                print(f'Error on candidate {candidate}:{e}')
        return notebooks

    def register_autotest(self):
        test_list = os.listdir("grading/testing/external_tests")
        test_list = [name.split('.')[0] for name in test_list if name.endswith(".py")]
        return test_list

    def run_autotests(self):
        test_results = {}
        sys.path.insert(0, f'{os.getcwd()}/grading/testing/external_tests')
        sys.path.insert(1, f'{os.getcwd()}/grading/testing/notebook_tests')
        for the_test in self.register_autotest():
            print(f"Now running test {the_test}")
            test_results[the_test]= self.run_single_test(the_test)
            test_results[the_test].to_csv(f'grading/scores/{the_test}.csv')

        self.test_results = test_results
        
        return self.test_results

    def run_single_test(self, the_test):
        results = None
        
        for candidate in tqdm(self.candidates):
            mod = __import__(the_test, fromlist=[the_test])
            tester = getattr(mod, the_test)
            result = tester(folder=f'{self.base_dir}/{candidate}').run_test()

            if results is None:
                results = pd.DataFrame(data=[result.values()],columns=result.keys())
                results['candidate'] = candidate
            else:
                formatted_result = pd.DataFrame(data=[result.values()],columns=result.keys())
                formatted_result['candidate'] = candidate
                results = pd.concat([results,formatted_result])
        
        return results.reset_index(drop=True)



    def run_notebooks(self,save_originals=None, kernel='python3', first_run=True):

        if first_run:
            try:
                results = pd.read_csv('grading/scores/nbta_runs.csv')
            except:
                results = pd.DataFrame(data=[], columns=['candidate','status','message'])

        ep = ExecutePreprocessor(kernel_name=kernel,
        allow_errors=True)

        for auth, notebook in tqdm(self.notebooks.items()):

            if (first_run and auth not in results.candidate.values) or (not first_run):
                passed,auth_errors = self.run_single_notebook(auth, notebook.modified_content.copy(),save_originals,ep)
                row = {'candidate': auth, 'status': passed, 'message': auth_errors}
                results = results.append(row, ignore_index = True)
                results.to_csv('grading/scores/nbta_runs.csv', index=False)
            
        self.runs_results = pd.read_csv('grading/scores/nbta_runs.csv')

        return self

    def run_single_notebook(self, auth, nb, save_originals, ep):
        super_path = os.getcwd()
        if save_originals is not None:
            for original_file in save_originals:
                try:
                    shutil.copy(f'{self.base_dir}/{auth}/{original_file}', f'{self.base_dir}/{auth}/nbta_{original_file}')
                except:
                    pass
        os.chdir(f'{super_path}/{self.base_dir}/{auth}')
        notebook_filename_out = f'{self.notebook_name}_marking.ipynb'

        try:
            out = ep.preprocess(nb, {'metadata': {'path': ''}})
        except CellExecutionError as e:
            print(e)
        finally:
            with open(notebook_filename_out, mode='w', encoding='utf-8') as f:
                nbf.write(nb, f)
        auth_errors = []
        passed = True
        for cell in nb['cells']:
            if cell['cell_type']=='code':
                for output in cell['outputs']:
                    if output['output_type'] == 'error':
                        auth_errors.append(f"{output['ename']}: {output['evalue']}")
                        passed = False

        os.chdir(super_path)
        return passed, auth_errors

    def generate_question_notebooks(self):
        questions = self.questions
        for index,question in enumerate(questions):
            if index+1 != len(questions):
                next_question = questions[index+1]
            else:
                next_question = None

            cells = []
            for auth, notebook in self.notebooks.items():
                cells = cells + notebook.yield_question_cells(question,next_question)
                  
            question_notebook = nbf.v4.new_notebook()    
            question_notebook['cells'] = cells

            with open(f'{question.name}_all_answers.ipynb', 'w') as f:
                nbf.write(question_notebook,f)



if __name__ == '__main__':
    with open('code_cells/final_cell.py') as f:
            lines = f.readlines()
    print(lines)



from datetime import datetime
import nbformat as nbf 
import os

def header_cell(author):
    # datetime object containing current date and time
    cells = []
    now = datetime.now()
    dt_string = now.strftime("on %d/%m/%Y at %H:%M:%S")

    source = [f'# MARKING NOTEBOOK [{author}]',
    f'### Generated {dt_string}']
    source = '\n'.join(source)

    cells.append(nbf.v4.new_markdown_cell(source=source))

    return cells

def header_code_cell(author):
    absolute_path = os.getcwd()

    source = ['%load_ext autoreload',
    '%autoreload 2',
    '',
    'import numpy as np',
    'import pandas as pd',
    'import matplotlib.pyplot as plt',
    'from nbta.grading import QuestionGrader, EstimatedMark',
    'import sys'
    '',
    '# Adding local test classes to the Python path:',
    f'sys.path.insert(0, f"../../grading/testing/notebook_tests")',
    f'sys.path.insert(1, f"../../grading/testing/external_tests")'
    ]
    
    source = '\n'.join(source)

    return nbf.v4.new_code_cell(source=source)
    
def footer_cell():
    lines = [f"nbta_test_style = QuestionGrader('style')",
    "nbta_estimated_mark = EstimatedMark()",
    f"nbta_private_feedback = QuestionGrader('Private_feedback_for_lecturer', feedback=True)"]
    return "\n".join(lines)



class NotAssigned():
    def __init__(self,expected_type=''):
        self.expected_type = expected_type
        return None


from bs4 import BeautifulSoup
import json
from os import listdir
from os.path import isfile, join

class htmlToIpnb:
    def __init__(self,folder='../raw_data/htmls'):
        self.folder=folder
        
    def convert_file(self, file_name):
        response = open(f"{self.folder}/{file_name}")
        text = response.read()

        soup = BeautifulSoup(text, 'lxml')
        dictionary = {'nbformat': 4, 'nbformat_minor': 1, 'cells': [], 'metadata': {}}
        for d in soup.findAll("div"):
            if 'class' in d.attrs.keys():
                for clas in d.attrs["class"]:
                    if clas in ["text_cell_render", "input_area"]:
                    # code cell
                        if clas == "input_area":
                            cell = {}
                            cell['metadata'] = {}
                            cell['outputs'] = []
                            cell['source'] = [d.get_text()]
                            cell['execution_count'] = None
                            cell['cell_type'] = 'code'
                            dictionary['cells'].append(cell)

                        else:
                            cell = {}
                            cell['metadata'] = {}

                            cell['source'] = [d.decode_contents()]
                            cell['cell_type'] = 'markdown'
                            dictionary['cells'].append(cell)
        notebook_name=f"{self.folder}/{file_name.strip('.html')}.ipynb"
        
        open(notebook_name, 'w').write(json.dumps(dictionary))
        print(f'Notebook {notebook_name} succesfully created')
        
        return None
    
    
    def convert(self):
        htmls = [f for f in listdir(self.folder) if f.split('.')[-1]=='html']
        for html in htmls:
            self.convert_file(html)
        
        print(f'Converted {len(htmls)} files.')
        
        return None

if __name__ == '__main__':
    htmlToIpnb().convert()
        



import pandas as pd
import numpy as np
import ipywidgets as widgets
from IPython.display import display, HTML
import os
from tqdm import tqdm

class NBTATest:
    def __init__(self,folder=None,**kwargs):
        self.results = kwargs
        self.folder = folder
        self.test_results = None

        if not os.path.exists('grades'):
                os.mkdir('grades')

    def run_test(self):
        pass

    def get_results(self):
        return self.test_results

class QuestionGrader:
    def __init__(self, question_name, options=None, auto_tests=None, feedback=False):
        self.base_dir = '../..'
        self.question_name = question_name
        self.auto_tests=auto_tests
        self.auto_tests_results={}
        self.feedback = feedback

        if not os.path.exists('grades'):
                os.mkdir('grades')

        if self.auto_tests is not None:
            self.run_autotests()

        if self.feedback:
            try:
                self.selected_pos_feedback = pd.read_csv(f'grades/nbta_pos_feedback_{self.question_name}.txt')
            except:
                self.selected_pos_feedback = None
            
            try:
                self.selected_neg_feedback = pd.read_csv(f'grades/nbta_neg_feedback_{self.question_name}.txt')
            except:
                self.selected_neg_feedback = None
            
            self.feedback = self.set_feedback()
        else:
            try:
                self.selected_values = pd.read_csv(f'grades/nbta_selection_{self.question_name}.csv')['options'].values
            except:
                self.selected_values = None
        
            if options is None:
                self.options = self.read_options(f'{self.base_dir}/grading/testing/notebook_tests/{self.question_name}.csv')
            else:
                self.options = options
        
            self.autoresults = self.set_autoresults_widget()
            self.widgets = self.set_widgets()

        self.display()

    def run_autotests(self):
        for name, test in self.auto_tests.items():
            res = test.run_test()
            res_df = pd.DataFrame.from_dict(res,orient='index').T
            res_df.to_csv(f'grades/nbta_score_{name}.csv', index=False)
            self.auto_tests_results[name] = res_df
            
    def read_options(self, path):
        return pd.read_csv(path).options.values
    
    def set_widgets(self):
        content = []
        for option in self.options:
            this_value = False
            if self.selected_values is not None:
                if option in self.selected_values:
                    this_value = True
                
            w = widgets.Checkbox(value=this_value,
                description=option,
                disabled=False,
                indent=False)
            content.append(w)
        return content
    
    def set_autoresults_widget(self):
        display_results = []
        if self.auto_tests is not None:
            for test_name, results in self.auto_tests_results.items():
                lines = [f'[{test_name}] => ']+ [f'{k}: {v}' for k,v in results.items()]
                display_results.append(','.join(lines))
        display_results = '\n'.join(display_results)

        autoresults = widgets.Textarea(
            value=display_results,
            placeholder=display_results,
            description='Results from automatic tests:',
            disabled=True)
        
        return autoresults


    def set_feedback(self):
        feedbacks = []
        if self.selected_pos_feedback is not None:
            with open(f'grades/nbta_pos_feedback_{self.question_name}.txt', 'r') as f:
                pos_feedback_text = '\n'.join(f.readlines())
        else:
            pos_feedback_text = ' '
        
        if self.selected_neg_feedback is not None:
            with open(f'grades/nbta_neg_feedback_{self.question_name}.txt', 'r') as f:
                neg_feedback_text = '\n'.join(f.readlines())
        else:
            neg_feedback_text = ' '

        feedbacks.append(widgets.Textarea(
            value=pos_feedback_text,
            placeholder=pos_feedback_text,
            description='POSITIVES:',
            disabled=False))

        feedbacks.append(widgets.Textarea(
            value=neg_feedback_text,
            placeholder=neg_feedback_text,
            description='NEGATIVES:',
            disabled=False))
        
        return feedbacks
    
    def display(self):
        if self.feedback:
            display(HTML(f'<h1 style="color:blue">{self.question_name.replace("_"," ")}</h1>'))
            display(HTML(f'<h4 style="color:red">Remember to be positive and constructive in your feedbacks!</h4>'))
            display(HTML(f'<p>Students are very sensitive to feedback, and will react to unfair statements. The goal is to provide useful information that will help students improve. When you write negative feedback, frame them as things that can be improved, and try to explain how. Balance the positive and the negative feedbacks, and give about 3-4 each positive & negative feedbacks per question. <br>Remember that your text will be added to more feedback from the automatic and manual tests, and is addressed to the student (so you can use "you").</p>'))
            for feedback in self.feedback:
                display(feedback)
        else:
            display(HTML(f'<h1 style="color:blue">{self.question_name.replace("_"," ")}</h1>'))

            if len(self.auto_tests_results.items()) > 0:
                display(HTML(f'<h2 style="color:teal">AUTOMATIC TESTS</h2>'))
            for test,df in self.auto_tests_results.items():
                display(HTML(f'<h3 style="color:purple">{test.upper()}</h3>'))
                for col in df.columns:
                    display(HTML(f'<b>{col}:</b>'))
                    display(HTML(f'{df[col].values}'))
        
            display(HTML(f'<h2 style="color:teal">MANUAL CHECKS</h2>'))
            for w in self.widgets:
                display(w)

        
    def values(self):
        values = {'values':[w.description for w in self.widgets if w.value]}
        return values

    def save_values(self):
        if self.feedback:
            with open(f'grades/nbta_pos_feedback_{self.question_name}.txt', 'w') as f:
                f.write(self.feedback[0].value)
            with open(f'grades/nbta_neg_feedback_{self.question_name}.txt', 'w') as f:
                f.write(self.feedback[1].value)
            print(f'Saved feedbacks for {self.question_name}')
        else:
            pd.Series(data=self.values().get('values'),
            dtype=object, 
            name='options').to_csv(f'grades/nbta_selection_{self.question_name}.csv', index=False)
            print(f'Saved options for {self.question_name}')



class EstimatedMark:
    def __init__(self, question_name='estimated_mark', options=None, auto_tests=None):
        self.base_dir = '../..'
        self.question_name = question_name
        self.grade_selector = self.set_grade_selector()
        self.display()
    
    def set_grade_selector(self):
        try:
            self.selected_option = pd.read_csv(f'grades/nbta_selection_{self.question_name}.csv')['options'].values.astype(str)
        except:
            self.selected_option = '2.1'

        est_grade = widgets.RadioButtons(
            options=['High 1st', '1st', '2.1', '2.2', '3rd', 'fail'],
            value=self.selected_option,
            description='Assessed score:',
            disabled=False)

        return est_grade
    
    def display(self):
        display(HTML(f"<h2>ESTIMATED MARK</h2>"))
        display(self.grade_selector)
        
    def save_values(self):
        pd.Series(data=[self.grade_selector.value], 
        name='options').to_csv(f'grades/nbta_selection_{self.question_name}.csv', index=False)
        print(f'Saved {self.question_name}')


class GradingSchema():
    def __init__(self,name,total_points,add_tests=None, ignore_feedbacks=[]):
        self.name = name
        self.total_points = total_points
        self.add_tests = add_tests
        self.markings = None
        self.style = None
        self.grades = None
        self.ignore_feedbacks = ignore_feedbacks

    def load_marks(self, folder, candidates):
        question_cols = pd.read_csv(f'grading/testing/notebook_tests/{self.name}.csv').options.values
        cols = ['candidate','additional_comments']+list(question_cols)
        marking = pd.DataFrame(data=[], columns=cols)

        for candidate in tqdm(candidates):
            init_values = [candidate,''] + [np.nan for _ in question_cols]
            marking.loc[marking.shape[0]+1] = init_values

            try:
                base_dir = f'{folder}/{candidate}/grades'
                values = pd.read_csv(f'{base_dir}/nbta_selection_{self.name}.csv')['options'].values
                with open(f'{base_dir}/nbta_feedback_{self.name}.txt') as f:
                    text = f.readlines()
                marking.loc[marking.shape[0],'additional_comments'] = "\n".join(text)
                marking.loc[marking.shape[0],question_cols] = False
                marking.loc[marking.shape[0],values] = True
                self.style = pd.read_csv(f'{base_dir}/nbta_selection_style.csv')['options'].values
                self.estimated_grade = pd.read_csv(f'{base_dir}/nbta_selection_estimated_grade.csv')['options'].values
            except Exception as e:
                print(f'Failed on candidate {candidate}: {e}')
        
        for bool_question in question_cols:
            marking.loc[:,bool_question] = marking.loc[:,bool_question].astype(bool)

        if self.add_tests is not None:
            for test in self.add_tests:
                test_results = pd.read_csv(f'grading/scores/{test}.csv')
                marking = marking.merge(test_results, on='candidate', how='outer')
        
        self.markings = marking.copy()
        return self

    def build_feedback(self, values, all_feedbacks):
        if values['not_answered'] == 1:
            return f'You did not answer {self.name}.'

        neg_feedbacks = []
        pos_feedbacks = []
        for f in all_feedbacks:
            if values[f] == 0:
                neg_feedbacks.append(f)
            else:
                pos_feedbacks.append(f)
        feedback_string = ''
        pos_feedback_string = ''
        neg_feedback_string = ''
        if len(pos_feedbacks)>0:
            pos_feedback_string = f'You did the following well in {self.name}:\n'
            pos_items = '\n'.join([f'    • {self.feedback_dict.get(f)}' for f in pos_feedbacks])
            pos_feedback_string = f'{pos_feedback_string}{pos_items}\n'
        if len(neg_feedbacks)>0:
            neg_feedback_string = f'Improvements in {self.name} would be possible for the following (either your did not do that, or not to the highest standard):\n'
            neg_items = '\n'.join([f'    • {self.feedback_dict.get(f)}' for f in neg_feedbacks])
            neg_feedback_string = f'{neg_feedback_string}{neg_items}\n'
        add_feedback = ''
        if values['additional_comments'] not in ['', ' ']:
            add_feedback = f'Additional feedback for this question: {values["additional_comments"]}\n\n'

        feedback_string = f'{pos_feedback_string}\n{neg_feedback_string}\n{add_feedback}'

        return feedback_string

    def get_feedback(self):
        feedback_options = pd.read_csv(f'grading/testing/notebook_tests/{self.name}.csv')
        self.feedback_dict = dict(zip(feedback_options.options.values,feedback_options.feedback.values))
        feedbacks = pd.DataFrame()
        feedbacks['candidate'] = self.markings.candidate
        feedbacks['feedback'] = '\n'

        ignore = self.ignore_feedbacks+['candidate', 'additional_comments','not_answered']

        cols = [c for c in self.markings.columns.values if c not in ignore]

        neg_feedbacks = [f for f in cols if f.split('_')[0] == 'neg']
        pos_feedbacks = [f for f in cols if f not in neg_feedbacks]

        values = self.markings.copy()
        values[neg_feedbacks] = abs(values[neg_feedbacks]-1)

        all_feedbacks = neg_feedbacks + pos_feedbacks

        print(f"Generating feedback for {self.name}:")
        for index, row in tqdm(values.iterrows()):
            feedbacks.loc[index,'feedback'] = f'{self.build_feedback(row,all_feedbacks)}'

        self.feedbacks = feedbacks

        return feedbacks

    def simple_grading(self, df):
        marks = pd.Series(name='marks', data=np.zeros(df.shape[0]))
        max_points = 0
        for col in df.columns.values:
            if df[col].dtype == bool:
                if (col.split('_')[0] == 'neg') or (col == 'not_answered'):
                    marks = marks + np.abs(df[col].values-1)
                else:
                    marks = marks + df[col].values
                max_points += 1
        mask = np.abs(df.not_answered-1)
        marks = marks.apply(lambda x: 0 if x<0 else x)
        marks = marks.values*mask
        self.max_points = max_points
        final_mark = marks/max_points*self.total_points
        print(f'Max points: {max_points}, Total mark:{self.total_points}, Mean mark±1SD:{np.mean(final_mark)}±{np.std(final_mark)}')
        return final_mark


    def calculate_grades(self):
        grades = self.simple_grading(self.markings)
        return grades
    
    def grade(self):
        grade_per_candidate = pd.DataFrame()
        grade_per_candidate['candidate'] = self.markings.candidate
        grade_per_candidate[self.name] = self.calculate_grades()
        self.grades = grade_per_candidate
        return self.grades



class Grader():
    def __init__(self, marker, schemas):
        self.marker = marker
        self.reset_grades()
        self.style_schema = GradingSchema('style',100)
        self.schemas = dict(zip([s.name for s in schemas],schemas))
        self.folder = self.marker.base_dir

    def make_feedback(self, pre='', post=''):
        self.grades['feedback'] = self.grades['candidate'].apply(lambda x: f'{pre}\n')
        
        for q in tqdm(self.schemas.keys()):
            current_feedback = self.grades[['candidate','feedback']].copy()
            current_feedback = current_feedback.merge(self.schemas.get(q).get_feedback(), on='candidate')
            current_feedback['feedback'] = current_feedback['feedback_x'] + current_feedback['feedback_y']
            self.grades['feedback'] = current_feedback['feedback'].copy()


        self.grades['feedback'] = self.grades['feedback'].apply(lambda x: f'{x}{post}')

        return self

    def reset_grades(self):
        self.grades = pd.DataFrame()
        self.grades['candidate'] = self.marker.candidates

    def grade(self, questions=None):
        self.reset_grades()

        if questions is None:
            questions = self.schemas.keys()
        elif not questions.isinstance(list):
            questions = [questions]
        
        self.grades['total'] = 0
        for question in questions:
            print(f"Grading question {question}")
            schema = self.schemas.get(question)
            schema.load_marks(self.folder, self.grades['candidate'].values)
            self.grades = self.grades.merge(schema.grade(), on = 'candidate')
            self.grades['total'] = self.grades['total'] + self.grades[schema.name]
        self.grades['total'] = np.round(self.grades['total'],0)

        print(f"Grading coding style")
        self.style_schema.load_marks(self.folder, self.grades['candidate'].values)
        style_results = self.style_schema.grade()
        style_results.columns = ['candidate','coding_style']
        self.grades = self.grades.merge(style_results, on = 'candidate')
        
        print(f"Gathering estimated marks")
        est_marks = pd.DataFrame()
        est_marks['candidate'] = self.grades['candidate']
        est_marks['estimated_marks']=np.nan
        for candidate in tqdm(self.grades['candidate'].values):
            mark = pd.read_csv(f'{self.folder}/{candidate}/grades/nbta_selection_estimated_grade.csv')
            est_marks.loc[est_marks.candidate==candidate,'estimated_marks']=mark.loc[0,'options']
        self.grades = self.grades.merge(est_marks, on = 'candidate')
        return self

    def marking_for_question(self, question):
        return self.schemas.get(question).markings

    def grading_for_question(self, question):
        return self.schemas.get(question).grades




from typing import DefaultDict
from termcolor import colored
import numpy as np
import pandas as pd
import pickle 
import numpy as np
import pandas as pd
from os import listdir
from wordcloud import WordCloud, STOPWORDS
from IPython.display import display
import matplotlib.pyplot as plt
from IPython.core.display import HTML
from termcolor import colored

class NBTest:
    def __init__(self,testName='test',**kwargs):
        self.testName = testName
        self.variables = kwargs
        self.results = []
        self.noProblems = True
    
    def setup(self):
        self.setupFromFile(f'{self.testName}.cvs')
        return self

    def setupFromFile(self, filename):
        testData = pd.read_csv(filename)

        for index, item in testData.iterrows():
            try:
                data = {'result':self.variable.get(item[0]),
                'expected':item[1],
                'error_message':item[2]}
                self.results.append(data)
            except:
                pass
        
        
    def run(self):
        pass
    
    def generateString(self, items):
        text=f'CHECKING AND SAVING VARIABLES FOR TEST "{self.testName}"'
        base_string = f"""
        \n======================================================================
        \n{colored(text, 'blue',attrs=['bold'])}"""
        
        substrings =  '\n----------------------------------------------------------------------\n'.join(items)
        return f"{base_string}{substrings}"
    
    def save_variable(self, variable, filename):
        outfile = open(f"answers/{filename}.pkl",'wb')
        data = [self.cid,variable]
        pickle.dump(data,outfile)
        outfile.close()
        
        return message


class CheckCode:
    def __init__(self,**kwargs):
        self.results = kwargs
        self.noProblems = True
        
    def checkAll(self):
        messages = [self.checkSetup(),
                    self.checkPartA(),
                    self.checkPartB(),
                    "\n======================================================================"]
        if self.noProblems:
            messages.append(colored('\nALL GOOD - PUSH TO GITHUB\n', 'green',attrs=['bold','underline']))
            messages.append(colored('\nRemember to run this cell again if you change your code\n', 'magenta',attrs=['bold','underline']))
        else:
            messages.append(colored('\nTHERE ARE SOME POTENTIAL PROBLEMS WITH YOUR CODE', 'magenta',attrs=['bold','underline']))
        return '\n'.join(messages)
        
    def checkSetup(self):
        cid = self.results['cid']
        self.cid = cid
        username = self.results['username']
        items = ['\n']
        
        items.append(self.test_type(cid,'cid','INTEGER','the setup section'))
        items.append(self.test_type(username,'GitHubUsername','STRING','the setup section'))
        
        return self.generateString('GENERAL SETUP',items)
    
    def checkPartA(self):
        items = ['\n']
        
        types = [(self.results['cdf_nb_missing'],'cdf_nb_missing', 'INTEGER', 'Question 1 (a)'),
        (self.results['cdf_std_mean'], 'cdf_std_mean', "FLOAT", 'Question 1 (b)'),
        (self.results['cdf_min_carb'], 'cdf_min_carb', "FLOAT", 'Question 1 (c)'), 
        (self.results['cdf_max_carb'], 'cdf_max_carb', "FLOAT", 'Question 1 (c)'),
        (self.results['num_columns'], 'num_columns', 'Numpy Array', 'Question 2'),
        (self.results['outliers_present'], 'outliers_present', 'STRING', 'Question 4')
        ]
        
        items = items + [self.test_type(var,name,typ,loc) for var,name,typ,loc in types]
        
        dfs = [(self.results['core_df'],'core_df', 'Questions 1-4'),
        (self.results['encoded_df'],'encoded_df', 'Question 5'),]
        items = items + [self.test_dataframe(var,name,loc) for var,name,loc in dfs]
        
        return self.generateString('PART A',items)
        return ''
    
    def checkPartB(self):
        types = [(self.results['final_model'],'final_model', 'LogisticRegression()', 'Question 6'),
                 (self.results['predictions'],'predictions', 'Numpy Array', 'Question 6')]
        
        items = ['\n'] + [self.test_type(var,name,typ,loc) for var,name,typ,loc in types]
        
        return self.generateString('PART B',items)
    
    def generateString(self, part, items):
        text=f'CHECKING AND SAVING VARIABLES IN {part}'
        base_string = f"""
        \n======================================================================
        \n{colored(text, 'blue',attrs=['bold'])}"""
        
        substrings =  '\n----------------------------------------------------------------------\n'.join(items)
        return f"{base_string}{substrings}"
    
    def save_variable(self, variable, filename):
        outfile = open(f"answers/{filename}.pkl",'wb')
        data = [self.cid,variable]
        pickle.dump(data,outfile)
        outfile.close()
    
    def test_dataframe(self, df, df_name, question):
        message = ''
        if type(df) == pd.DataFrame:
            message = f"A Pandas DataFrame named {df_name} exists: {colored('PASSED', 'green',attrs=['bold','underline'])}"       
        else:
            problem = f"{df_name} is not a pandas DataFrame: {colored('FAILED', 'red',attrs=['bold','underline'])}"
            suggestion = f'SUGGESTION:Check in "{question}" that you have saved your {df_name} as a DataFrame (instead of e.g. a Numpy array)'
            suggestion = colored(suggestion, 'yellow')
            message = '\n'.join([problem, suggestion])
            self.noProblems = False
        try:
            self.save_variable(df,df_name)
            message = f"{message}\n{colored(f'Content of {df_name} correctly saved', 'green')}" 
        except:
            message = f"{message}\n{colored('Unable to save data from the dataframe', 'red')}" 
        return message
            
    def test_type(self, variable, variable_name, expected_type, question):
        type_to_str = {
            int:'INTEGER',
            str:'STRING',
            float:'FLOAT',
            object:'OBJECT',
            np.int64:'INTEGER',
            np.float64:'FLOAT',
            np.ndarray: 'Numpy Array',
            LogisticRegression:'LogisticRegression()'
        }
        
        variable_type = type_to_str[type(variable)]
        message = ''
        suggestion = ''
        
        if variable_type == expected_type:
            message = f"{variable_name} exists and is of type {variable_type}: {colored('PASSED', 'green',attrs=['bold','underline'])}"
        else:
            problem = f"{variable_name} is of type {variable_type}: {colored('FAILED', 'red',attrs=['bold','underline'])}"
            suggestion = f'SUGGESTION:Check in "{question}" that the variable {variable_name} is of type {expected_type}'
            suggestion = colored(suggestion, 'yellow')
            message = '\n'.join([problem, suggestion])
            self.noProblems = False
            
        try:
            self.save_variable(variable,variable_name)
            suggestion = colored(f"Variable {variable_name} was saved as a pickle file", 'green')
            message = '\n'.join([message, suggestion])
        except Exception as e:
            message = f'{e}'
            suggestion = colored(f"Unable to save variable {variable_name} as a pickle file", 'red')
            message = '\n'.join([message, suggestion])
        
        return message

class QuizCheck():
    '''
    QuizCheck offers a class to provide visual feedbacks on small quizes conducted through Google Forms.
    '''
    def __init__(self, path, answer_file=None):
        self.path = path
        self.answers = self.build_dataframe(path)
        self.correct_answers = False
        if answer_file:
            self.correct_answers = pd.read_excel(f'{path}/{answer_file}')
      
        self.questions = self.get_questions(self.answers)
        display(HTML(f'<h2>Number of questions:{len(self.questions)}</h2>'))
    
    def build_dataframe(self, path):
        forms = [form for form in listdir(path) if form.split('.')[-1]=="csv"]
        forms.sort()
        df = pd.read_csv(f'{path}/{forms[0]}')
        for form in forms[1:]:
            df = pd.concat([df,pd.read_csv(f'{path}/{form}')])
        return df
     
    def get_questions(self, df):
        allColumns = list(df.columns)
        allColumns.remove('Timestamp')
        allColumns.remove('Total score')
        return [question for question in allColumns if question.find(' [Score]') ==-1 if question.find(' [Feedback]') ==-1]
    
    def split_answers(self, answer):
        return answer.split(';')
    
    def num_to_words(self, num):
        dic = {
            0:'zero',
            1:'one',
            2:'two',
            3:'three',
            4:'four',
            5:'five',
            6:'six',
            7:'seven',
            8:'eight',
            9:'nine'
        }

        try:
            num=int(float(num))
            return dic.get(num, 'many')
        except:
            return num


    def draw_WordCloud(self, question, image):
        fig, ax = plt.subplots(1,1, figsize=(15,8))
        ans = self.answers[question].dropna()
        text = " ".join(self.num_to_words(answer) for answer in ans.astype(str))
    
        # lower max_font_size, change the maximum number of word and lighten the background:
        wordcloud = WordCloud(max_font_size=50, max_words=100, background_color="white", collocations=True).generate(text)

        plt.imshow(wordcloud, interpolation="bilinear")
        plt.axis("off")
        fig.tight_layout(pad=6.0)
        #plt.rc('xtick',labelsize=10)
        #plt.rc('ytick',labelsize=30)
        plt.show()

    def format_numerical_answer(self, ans):
        ans = ans.dropna()
        nb_responses = len(ans)
    
        all_ans = []

        for value in ans.values:
            all_ans = all_ans+self.split_answers(value)
        ans = pd.Series(all_ans)

        data = ans.value_counts().sort_index(ascending=True)/nb_responses*100

        return data

    def draw_bars(self, question,  image):
        _, ax = plt.subplots(1,1, figsize=(9,9))
        ans = self.answers[question].dropna()
        nb_responses = len(ans)
    
        all_ans = []
        for value in ans.values:
            all_ans = all_ans+self.split_answers(value)
        ans = pd.Series(all_ans)
    
        data = ans.value_counts().sort_index(ascending=True)/nb_responses*100

        height = data.values
        x_pos = np.arange(len(height))
    
        # Create colors
        colors = []
        for bar in data.index:
            if bar in self.correct_answers[question].values:
                colors.append('green')
            else:
                colors.append('blue')
        
        # Create bars
        plt.bar(x_pos, height, color=colors)

        # Create names on the x-axis
        plt.xticks(x_pos, x_pos)
        plt.yticks(list(range(0,110, 10)))
        plt.ylabel('Percentage Answers', fontsize=14)
        plt.xlabel('Answer Number', fontsize=14)
        
        # Show graph
        plt.show()
    
        for i,label in enumerate(data.index):
            color=(0,0,0)
            if label in self.correct_answers[question].values:
                color= "green"   
            display(HTML(f'<font color={color}>{i}:{label}</font>'))
        return ax

    def answer(self, question_nb, correct_answers=[], image=None):
        question = self.questions[question_nb]
        display(HTML(f'<h2>{question}</h2>'))
        if self.answers[f'{question} [Score]'].dropna()[0][0] == '-':
            self.draw_WordCloud(question, image)
        else:
            self.draw_bars(question, image)
        display(HTML(f'<h3> <font color="blue">{" "*100}</font></h3>'))
        display(HTML(f'<h4> <font color="grey">{"_"*100}</font></h4>'))
        display(HTML(f'<h3> <font color="blue">{" "*100}</font></h3>'))
        plt.show()

    def print_answers(self):
        for nbq in range(len(self.questions)):
            self.answer(nbq)
    
    def get_percentage_right(self, given_answers):
        for question in self.questions:
            correct_answers = []
            if self.answers[f'{question} [Score]'].dropna()[0][0] == '-':
                pass
            else:
                data = self.format_numerical_answer(given_answers[question])
                correct_answers = correct_answers+[data[ans] for ans in data.index if ans in self.correct_answers[question].values]
        
        return sum(correct_answers)/len(correct_answers)
    
    def compare_scores(self, other_years=[]):
        bars = [self.get_percentage_right(self.answers)]
        labels = ['2022']
        colors = ['green']

        if len(other_years)!=0:
            for year in other_years:
                other_answers = self.build_dataframe(f'{self.path}/{year}')
                bars.append(self.get_percentage_right(other_answers))
                labels.append(str(year))
                colors.append('blue')

        _, ax = plt.subplots(1,1, figsize=(9,9))
        
        plt.bar(labels, bars, color=colors)

        plt.ylabel('Percentage Answers', fontsize=14)
        plt.xlabel('Year', fontsize=14)
        
        # Show graph
        plt.show()
            